//  This file is part of ff3d - http://www.freefem.org/ff3d
//  Copyright (C) 2008 Stphane Del Pino
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  $Id: DGFunctionBase.hpp,v 1.1 2008/01/31 21:38:15 delpinux Exp $

#ifndef DG_FUNCTION_BASE_HPP
#define DG_FUNCTION_BASE_HPP

#include <ScalarFunctionBase.hpp>

#include <ScalarDegreeOfFreedomPositionsSet.hpp>
#include <DegreeOfFreedomSetManager.hpp>

#include <ScalarDiscretizationTypeDG.hpp>

#include <Vector.hpp>

class Mesh;

/**
 * @file   DGFunctionBase.hpp
 * @author Stphane Del Pino
 * @date   Thu Jan 31 21:53:23 2008
 * 
 * @brief base class for discontinuous finite element functions
 */
class DGFunctionBase
  : public ScalarFunctionBase
{
protected:
  ConstReferenceCounting<Mesh>
  __baseMesh;			/**< reference to the mesh through the
				   Mesh (basis) class */

  ConstReferenceCounting<ScalarDegreeOfFreedomPositionsSet>
  __dofPositionsSet;		/**< degree of freedom position set */

  Vector<real_t> __values;	/**< vector of values */

  const ScalarDiscretizationTypeDG
  __discretizationType;		/**< type of discretization*/

  /** 
   * output of the DGFunction
   * 
   * @param os output stream
   * 
   * @return os
   */
  std::ostream& __put(std::ostream& os) const
  {
    os << "{dg-" << ScalarDiscretizationTypeBase::name(__discretizationType) << '}';
    return os;
  }

  /**
   * this value is returned when function is evaluated outside the
   * mesh.
   * 
   * @note It is required to allow expression like @f$ f+a @f$ (where
   * @f$ f @f$ is a dg-function and @f$ a @f$ a constant) to be
   * simplifed (DG functions are considered to be 0 outside the mesh
   * by convention).
   */
  real_t __outsideValue;

public:
  /** 
   * set outside value
   * 
   * @param value outside value
   */
  void setOutsideValue(const real_t& value)
  {
    __outsideValue = value;
  }

  /** 
   * Access to the outside value
   * 
   * @return __outsidValue
   */
  const real_t& outsideValue() const
  {
    return __outsideValue;
  }

  /** 
   * Checks if the function can be simplified
   * 
   * @return true
   */
  bool canBeSimplified() const
  {
#warning set to false since not implemented yet
    return false;
  }

  /** 
   * Check if a dg function is of the same type
   * 
   * @param f given function
   * 
   * @return true if discretization type and mesh are the same
   */
  bool hasSameType(const DGFunctionBase& f) const
  {
    return ((__discretizationType.type() == f.__discretizationType.type())
	    and (__baseMesh == f.__baseMesh));
  }

  /** 
   * Access to the mesh through its base class
   * 
   * @return __baseMesh
   */
  ConstReferenceCounting<Mesh> baseMesh() const
  {
    return __baseMesh;
  }

  /** 
   * The type of the finite elememt function
   * 
   * @return __discretizationType
   */
  const ScalarDiscretizationTypeBase::Type& discretizationType() const
  {
    return __discretizationType.type();
  }

  /** 
   * Read-only access to the set of values
   * 
   * @return __values
   */
  const Vector<real_t>& values() const
  {
    return __values;
  }

  /** 
   * access to the value at the ith degree of freedom
   * 
   * @param i number of the degree of freedom
   * 
   * @return __values[i]
   */
  inline real_t& operator[](const size_t& i)
  {
    return __values[i];
  }

  /** 
   * Read only Access to the value at the ith degree of freedom
   * 
   * @param i number of the degree of freedom
   * 
   * @return __values[i]
   */
  inline const real_t& operator[](const size_t& i) const
  {
    return __values[i];
  }

  /** 
   * Affects the vector values
   * 
   * @param values new values
   */
  void operator=(const Vector<real_t>& values)
  {
    ASSERT (values.size() == __values.size());
    __values = values;
  }

  /** 
   * Affects a function to a DGFunction
   * 
   * @param f original function
   */
  virtual void operator=(const ScalarFunctionBase& f) = 0;

  /** 
   * Returns the gradient of a DGFunction at position @a x
   * 
   * @param x position of evaluation
   * 
   * @return @f$ \nabla f (x) @f$
   */
  virtual TinyVector<3,real_t>
  gradient(const TinyVector<3,real_t>& x) const = 0;

  /** 
   * Constructor
   * 
   * @param mesh the mesh supporting the function
   * @param discretizationType the type of finite element discretization
   */
  DGFunctionBase(const Mesh* mesh,
		 const ScalarDiscretizationTypeDG& discretizationType)
    : ScalarFunctionBase(ScalarFunctionBase::dgfunction),
      __baseMesh(mesh),
      __dofPositionsSet(DegreeOfFreedomSetManager::instance().getDOFPositionsSet(*mesh,discretizationType)),
      __values(__dofPositionsSet->number()),
      __discretizationType(discretizationType),
      __outsideValue(0)
  {
    ;
  }

  /** 
   * Destructor
   * 
   */
  virtual ~DGFunctionBase()
  {
    DegreeOfFreedomSetManager::instance().unsubscribe(__dofPositionsSet);
  }
};

#endif // DG_FUNCTION_BASE_HPP
